/*******************************************************
 NAME:		KF_Equations.h
 AUTHOR:	Luis Montesano
 DATE:		06-04-04
 DESCRIPTION:
	Contains the header of the kalman filter implementation
	for tracking people (or moving objects) based on xy 
	observations of the positions of the objects.

********************************************************/

#include <math.h>

#include "KF_Equations.h"

#define DIM_KF 4

typedef struct {
	float s2vx;		// x-acceleration noise
	float s2vy;		// y-acceleration noise
	float os2x;		// x-observation noise
	float os2y;		// y-observation noise
} TConfiguration;

static TConfiguration configuration;
static MATRIX F, Q, H, R; 

void initKFModule(float s2vx, float s2vy, float os2x, float os2y) {
/* Initializes the kf module (motion and measurement models) */


// motion model: F and Q
configuration.s2vx=s2vx; configuration.s2vy=s2vy;
configuration.os2x=os2x; configuration.os2y=os2y;

identity_matrix(&F,DIM_KF);

initialize_matrix(&Q,DIM_KF, DIM_KF);
MDATA(Q,2,2)=configuration.s2vx;
MDATA(Q,3,3)=configuration.s2vy;

// measurement model: H, R
/* obseravations only give position information */
initialize_matrix(&H,2,4);
MDATA(H,0,0)=1;
MDATA(H,1,1)=1;   

// Observation noise
initialize_matrix(&R,2,2);
MDATA(R,0,0)=configuration.os2x;
MDATA(R,1,1)=configuration.os2y;

}




/**************************************************
  Creates a new kalman filter in the given position 

 [x,P]=initializeKF()

 INPUT:
   pos: position
 OUTPUT:
   kf: [x,P] state vector & covariance matrix 
***************************************************/

void initializeKF(Tpf const *pos, Tkf *kf) {

initialize_vector(&(kf->x),DIM_KF);
kf->x.data[0]=pos->x;
kf->x.data[1]=pos->y;

initialize_matrix(&(kf->P), DIM_KF, DIM_KF);
MDATA(kf->P,0,0)=configuration.os2x;
MDATA(kf->P,1,1)=configuration.os2y;
MDATA(kf->P,2,2)=configuration.s2vx;
MDATA(kf->P,3,3)=configuration.s2vy;

}


/****************************************************
  Kalman filter prediction step 

  int predKFPT(Tkf *inkf, double period, Tkf *predkf)

 INPUT:
   inkf: state vector x={x y vx vy} & P: covariance matrix   
   period: period

 OUTPUT:
   predkf: prediction of x & P   

 Cte velocity Kalman Filter
******************************************************/
void predKFPT(Tkf const *inkf, float period, Tkf *predkf){

MATRIX tmp;
/* cte velocity model, no action*/
MDATA(F,0,2)=period;
MDATA(F,1,3)=period;

multiply_matrix_vector(&F,&(inkf->x), &(predkf->x));

multiply_matrix_matrix_t_matrix(&F, &(inkf->P), &F, &tmp);
add_matrix(&tmp, &Q, &(predkf->P));
}


/******************************************************
 Kalman filter update step 

  int updateKFPT(Tkf predkf, Tpf z, Tkf updkf);

  [x,P]=updateKFPT(x_, P_, z)

 INPUT:
   x_: prediction of the state vector x={x y vx vy}
   P: covariance matrix of the prediction

 OUTPUT:
   x: state vector
   P: Covariance matrix
*******************************************************/

void updateKFPT(Tkf const *predkf, Tpf const *z, Tkf *updkf) {

VECTOR v,vtmp;
MATRIX tmp,itmp,S,W;
	
  //v=z-H*x_;
  multiply_matrix_vector(&H,&(predkf->x),&vtmp);
  initialize_vector(&v,2);
  VDATA(v,0)=z->x-VDATA(vtmp,0);
  VDATA(v,1)=z->y-VDATA(vtmp,1);

  //S=H*P_*H'+R;
  multiply_matrix_matrix_t_matrix(&H, &(predkf->P), &H, &tmp);
  add_matrix(&tmp, &R, &S);

  //W=P_*H'*inv(S);
  multiply_matrix_t_matrix(&(predkf->P), &H, &tmp);
  inverse_matrix (&S, &itmp);
  multiply_matrix(&tmp,&itmp,&W);

  //x=x_+W*v;
  multiply_matrix_vector(&W,&v,&vtmp);
  add_vector(&(predkf->x), &vtmp, &(updkf->x));
  
  //P=P_-W*S*W';
  multiply_matrix_matrix_t_matrix(&W, &S, &W, &tmp);
  substract_matrix(&(predkf->P), &tmp, &(updkf->P));
}


/* uncertain motion */
int moveKF(Tsc const *motion,  Tkf const *kfs, Tkf *movedkfs){

/*float cost,sint,sintx,costy,sinty;
float tmp1,tmp2, vxa,vya;
MATRIX fx;
int I;*/
	return 0;
}
/*
  nKF=KF;
  
  //
  Pmotion=odometryModel(sqrt(motion(1)^2+motion(2)^2), abs(motion(3)));
  //


  cost=cos(motion->th);
  sint=sin(motion->th);
  costx=cost*motion->x;
  sintx=sint*motion->x;
  costy=cost*motion->y;
  sinty=sint*motion->y;

  initialize_matrix(fx,DIM_KF,7);
  MDATA(fx,0,0)=cost; MDATA(fx,0,1)=sint;
  MDATA(fx,0,4)=-cost; MDATA(fx,0,5)=-sint;
  MDATA(fx,1,0)=-sint; MDATA(fx,1,1)=cost;
  MDATA(fx,1,4)=sint; MDATA(fx,1,5)=-cost;
  MDATA(fx,2,2)=cost;MDATA(fx,2,3)=sint;
  MDATA(fx,3,2)=-sint;MDATA(fx,3,3)=cost;
   
  tmp1=-sinty - costx;
  tmp2=-costy +sintx;
  
  
  for (I=0;I<numKF;I++) {   
	xa=VDATA(kfs(I)->x,0);
    ya=VDATA(kfs(I)->x,1);
    vxa=VDATA(kfs(I)->x,2);
    vya=VDATA(kfs(I)->x,3);
    VDATA(movedkfs(I)->x,0)=cost*xa + sint*ya +  tmp1;
    VDATA(movedkfs(I)->x,1)=-sint*xa + cost*ya +  tmp2;
    VDATA(movedkfs(I)->x,2)=cost*vxa + sint*vya;
    VDATA(movedkfs(I)->x,3)=-sint*vxa + cost*vya;

    covP=[KF.kf(I).P zeros(4,3); zeros(3,4) Pmotion];
    
	MDATA(fx,6,0)=-costx+sinty-sint*xa+cost*ya;
    MDATA(fx,6,1)= sinty+costx-cost*xa-sint*ya;
    MDATA(fx,6,2)= -vxa*sint + vya*cost;
    MDATA(fx,6,3)= -vxa*cost - vya*sint;

	multiply_matrix_matrix_t_matrix(&fx, &covP, &fx, &(movedkfs(I)->P));    
  }

}




function Cov_odo_k1=odometryModel(dtheta, sq_dist)
% This function models the odometry error
%
% INPUTS: dtheta orientation variation
%         sq_dist square of the distance 
%
% The odometry error depends on the distance between the two locations
% and on the difference of orientation
% The error is assumed to be lower when the robot moves on a straight line
% and increases when it turns. Error on x and y are correlated.

turn_radius = 60000;

dist2=sqrt(sq_dist * (1.0 + 0.11 * dtheta/pi));

% errors due to distance
sxr = 0.1 * dist2;
syr = sq_dist/(2*turn_radius);
sthr = 2*dist2 / turn_radius;

% errors due to orientation
sxg = dtheta/pi * 4.4;
syg = dtheta/pi * 4.4;
sthg = 0.3 * dtheta;

Cov_odo_k1 = [sxr^2+sxg^2 0 0; 0 syr^2+syg^2 0.75*(syr * sthr); 0 0.75*(syr * sthr) sthr^2+sthg^2];

*/

void printFilter(Tkf *kf, char *msg){

	char msg1[10]="state: ";
	char msg2[15]="covariance: ";

	printf(msg);
	print_vector (msg1, &(kf->x));
	print_matrix (msg2, &(kf->P));
}
